/*
 * optimized local transport for VirtualGL designed for use with Bumblebee
 *
 * (c) 2012 Andreas Steinmetz
 *
 * This code is put in the public domain provided that the above header
 * is not modified or removed.
 */

#include <pthread.h>
#include <semaphore.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <dlfcn.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/extensions/XShm.h>
#include <X11/extensions/shmproto.h>
#include <rrtransport.h>

#define POOLSIZE 8

#define MIN(a,b) (a<b?a:b)

typedef struct _frame
{
	struct _frame *next;
	int used;
	int sync;
	sem_t sem;
	RRFrame rf;
} FRAME;

typedef struct
{
	int initialized;
	int shm;
	struct _ctx *ctx;
	FRAME *f;
	Display *dpy;
	XImage *xi;
	Drawable d;
	Pixmap pm;
	GC xgc;
	XShmSegmentInfo shminfo;
} WORKDATA;

typedef struct _ctx
{
	int notify;
	int memwait;
	int useshm;
	int format;
	Display *dpy;
	FakerConfig *fconfig;
	FRAME *pool;
	FRAME *head;
	FRAME *tail;

	void *xlib;
	int (*YCloseDisplay)(Display*);
	int (*YCopyArea)(Display*,Drawable,Drawable,GC,int,int,unsigned int,
		unsigned int,int,int);
	GC (*YCreateGC)(Display*,Drawable,unsigned long,XGCValues*);
	XImage *(*YCreateImage)(Display*,Visual*,unsigned int,int,int,char*,
		unsigned int,unsigned int,int,int);
	Pixmap (*YCreatePixmap)(Display*,Drawable,unsigned int,unsigned int,
		unsigned int);
	int (*YDestroyImage)(XImage *ximage);
	int (*YFlush)(Display*);
	int (*YFreeGC)(Display*,GC);
	int (*YFreePixmap)(Display*,Pixmap);
	Status (*YGetWindowAttributes)(Display*,Window,XWindowAttributes*);
	void (*YLockDisplay)(Display*);
	Display *(*YOpenDisplay)(_Xconst char*);
	int (*YPutImage)(Display*,Drawable,GC,XImage*,int,int,int,int,
		unsigned int,unsigned int);
	XErrorHandler (*YSetErrorHandler)(XErrorHandler);
	int (*YSync)(Display*,Bool);
	void (*YUnlockDisplay)(Display*);

	void *xext;
	Bool (*YShmAttach)(Display*,XShmSegmentInfo*);
	XImage *(*YShmCreateImage)(Display*,Visual*,unsigned int,int,char*,
		XShmSegmentInfo*,unsigned int,unsigned int);
	Bool (*YShmDetach)(Display*,XShmSegmentInfo*);
	Bool (*YShmPutImage)(Display*,Drawable,GC,XImage*,int,int,int,int,
		unsigned int,unsigned int,Bool);
	Bool (*YShmQueryExtension)(Display*);

	pthread_t h;
	Window win;
	pthread_mutex_t mtx;
	sem_t sem;
	sem_t msg;
	sem_t mem;
	FRAME fbuffer[POOLSIZE];
	WORKDATA w;
} CTX;

static char *errmsg="";
static unsigned long serial;
static int error;
static XErrorHandler prev;

static  int errhandler(Display *dpy, XErrorEvent *e)
{
	if(e->serial==serial&&e->minor_code==X_ShmAttach&&
		e->error_code==BadAccess)error=1;
	else if(prev&&prev!=errhandler)return prev(dpy, e);
	return 0;
}

static int getmode(CTX *ctx)
{
	XWindowAttributes attr;
	XImage *xi;
	int mode=-1;
	int red;
	int green;
	int blue;
	int alpha;
	int bgr;
	int ps;
	int i;

	if(!ctx->YGetWindowAttributes(ctx->dpy,ctx->win,&attr))goto fail1;

	if(!(xi=ctx->YCreateImage(ctx->dpy,attr.visual,attr.depth,ZPixmap,0,
		NULL,1,1,8,0)))goto fail1;

	switch(xi->red_mask)
	{
	case 0xff0000:
		red=4;
		break;
	case 0x00ff00:
		red=2;
		break;
	case 0x0000ff:
		red=1;
		break;
	default:red=0;
		break;
	}

	switch(xi->green_mask)
	{
	case 0xff0000:
		green=4;
		break;
	case 0x00ff00:
		green=2;
		break;
	case 0x0000ff:
		green=1;
		break;
	default:green=0;
		break;
	}

	switch(xi->blue_mask)
	{
	case 0xff0000:
		blue=4;
		break;
	case 0x00ff00:
		blue=2;
		break;
	case 0x0000ff:
		blue=1;
		break;
	default:blue=0;
		break;
	}

	ps=xi->bits_per_pixel>>3;

	if((red|green|blue)!=7||green!=2||ps<3)goto fail2;

	if(xi->byte_order==LSBFirst&&red==4)bgr=1;
	else if(xi->byte_order==MSBFirst&&red==1)bgr=1;
	else bgr=0;

	if(xi->byte_order==MSBFirst&&ps==4)alpha=1;
	else alpha=0;

	for(i=0;i<RRTRANS_FORMATOPT;i++)if(rrtrans_ps[i]==ps&&
		rrtrans_bgr[i]==bgr&&rrtrans_afirst[i]==alpha)
	{
		mode=i;
		break;
	}

fail2:	ctx->YDestroyImage(xi);
fail1:	return mode;
}

static void fbremove(WORKDATA *w)
{
	if(w->pm)
	{
		w->ctx->YFreePixmap(w->dpy,w->pm);
		w->pm=0;
	}
	if(w->xi)
	{
		if(w->xi->data&&!w->shm)
		{
			free(w->xi->data);
			w->xi->data=NULL;
		}
		w->ctx->YDestroyImage(w->xi);
	}
	if(w->shm)
	{
		w->ctx->YShmDetach(w->dpy,&w->shminfo);
		w->ctx->YSync(w->dpy,False);
		shmdt(w->shminfo.shmaddr);
	}
	if(w->xgc)w->ctx->YFreeGC(w->dpy,w->xgc);
}

static int fbsetup(WORKDATA *w)
{
	XWindowAttributes attr;

	if(w->initialized)
	{
		if(w->f->rf.w!=w->xi->width||w->f->rf.h!=w->xi->height)
		{
			fbremove(w);
			w->initialized=0;
		}
		else return 0;
	}

	w->pm=0;
	w->xi=NULL;
	w->shm=0;
	w->xgc=0;

	if(!w->ctx->YGetWindowAttributes(w->dpy,w->d,&attr))goto fail;

	if(w->ctx->useshm)
	{
		if(w->ctx->YShmQueryExtension(w->dpy))
		{
			w->shminfo.shmid=-1;
			w->shminfo.shmaddr=NULL;

			if(!(w->xi=w->ctx->YShmCreateImage(w->dpy,attr.visual,
				attr.depth,ZPixmap,NULL,&w->shminfo,w->f->rf.w,
					w->f->rf.h)))
			{
				w->ctx->useshm=0;
				goto noshm;
			}

			if((w->shminfo.shmid=shmget(IPC_PRIVATE,
				w->xi->bytes_per_line*w->xi->height+1,
				IPC_CREAT|0777))==-1)
			{
				w->ctx->useshm=0;
				w->ctx->YDestroyImage(w->xi);
				goto noshm;
			}

			if((w->shminfo.shmaddr=w->xi->data=(char *)shmat(
				w->shminfo.shmid,0,0))==(char *)-1)
			{
				w->ctx->useshm=0;
				w->ctx->YDestroyImage(w->xi);
				shmctl(w->shminfo.shmid,IPC_RMID,0);
				goto noshm;
			}

			w->shminfo.readOnly=False;

			w->ctx->YLockDisplay(w->dpy);
			w->ctx->YSync(w->dpy,False);
			prev=w->ctx->YSetErrorHandler(errhandler);
			error=0;
			serial=NextRequest(w->dpy);
			w->ctx->YShmAttach(w->dpy,&w->shminfo);
			w->ctx->YSync(w->dpy,False);
			w->ctx->YSetErrorHandler(prev);
			w->ctx->useshm=(error?0:1);
			w->ctx->YUnlockDisplay(w->dpy);

			shmctl(w->shminfo.shmid,IPC_RMID,0);

			if(!w->ctx->useshm)
			{
				w->ctx->YDestroyImage(w->xi);
				shmdt(w->shminfo.shmaddr);
				goto noshm;
			}

			w->shm=1;
		}
		else
		{
			w->ctx->useshm=0;
			goto noshm;
		}
	}
	else
	{
noshm:		if(!(w->pm=w->ctx->YCreatePixmap(w->dpy,w->d,attr.width,
			attr.height,attr.depth)))goto fail;

		if(!(w->xi=w->ctx->YCreateImage(w->dpy,attr.visual,attr.depth,
			ZPixmap,0,NULL,w->f->rf.w,w->f->rf.h,8,0)))goto fail;

		if(!(w->xi->data=(char *)malloc(w->xi->bytes_per_line*
			w->xi->height+1)))goto fail;
	}

	if(w->xi->width!=w->f->rf.w||w->xi->height!=w->f->rf.h)goto fail;

	if(!(w->xgc=w->ctx->YCreateGC(w->dpy,w->pm?w->pm:w->d,0,NULL)))
		goto fail;

	w->initialized=1;

	return 0;

fail:	fbremove(w);
	return -1;
}

static void cleaner(void *data)
{
	WORKDATA *w=(WORKDATA *)data;

	if(w->dpy)
	{
		fbremove(w);
		w->ctx->YCloseDisplay(w->dpy);
	}
}

static void *worker(void *data)
{
	CTX *ctx=(CTX *)data;
	WORKDATA *w=&ctx->w;
	int width;
	int height;
	int ps;
	int srcpitch;
	int dstpitch;
	int i;
	unsigned char *src;
	char *dst;
	sigset_t set;

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

	sigfillset(&set);
	pthread_sigmask(SIG_BLOCK,&set,NULL);

	memset(w,0,sizeof(WORKDATA));

	w->ctx=ctx;
	w->dpy=ctx->YOpenDisplay(DisplayString(ctx->dpy));
	w->d=ctx->win;

	pthread_cleanup_push(cleaner,w);

	while(1)
	{
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		sem_wait(&ctx->sem);
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

		pthread_mutex_lock(&ctx->mtx);
		if(!ctx->head)
		{
			pthread_mutex_unlock(&ctx->mtx);
			continue;
		}
		w->f=ctx->head;
		ctx->head=w->f->next;
		if(!ctx->head)ctx->tail=NULL;
		pthread_mutex_unlock(&ctx->mtx);

		if(!w->dpy)goto skip;
		if(fbsetup(w))goto skip;

		width=MIN(w->f->rf.w,w->xi->width);
		height=MIN(w->f->rf.h,w->xi->height);

		if(!width||!height)goto skip;

		srcpitch=w->f->rf.pitch;
		src=w->f->rf.bits+(w->f->rf.h-1)*srcpitch;
		dstpitch=w->xi->bytes_per_line;
		dst=w->xi->data;
		ps=width*rrtrans_ps[ctx->format];

		for(i=0;i<height;i++,src-=srcpitch,dst+=dstpitch)
			memcpy(dst,src,ps);

		if(w->shm)ctx->YShmPutImage(w->dpy,w->d,w->xgc,w->xi,
			0,0,0,0,width,height,False);
		else
		{
			ctx->YPutImage(w->dpy,w->pm,w->xgc,w->xi,
				0,0,0,0,width,height);
			ctx->YCopyArea(w->dpy,w->pm,w->d,w->xgc,
				0,0,width,height,0,0);
		}
		ctx->YFlush(w->dpy);
		ctx->YSync(w->dpy,False);

skip:		pthread_mutex_lock(&ctx->mtx);
		w->f->next=ctx->pool;
		ctx->pool=w->f;
		if(w->f->sync)sem_post(&w->f->sem);
		if(ctx->notify)sem_post(&ctx->msg);
		if(ctx->memwait)sem_post(&ctx->mem);
		pthread_mutex_unlock(&ctx->mtx);
	}

	pthread_cleanup_pop(1);

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	pthread_exit(NULL);
}

const char *RRTransGetError(void)
{
	return errmsg;
}

void *RRTransInit(Display *dpy,Window win,FakerConfig *fconfig)
{
	int i;
	CTX *ctx;
	char *env;

	if(!(ctx=malloc(sizeof(CTX))))
	{
		errmsg="RRTransInit: malloc failure";
		goto fail1;
	}

	ctx->dpy=dpy;
	ctx->win=win;
	ctx->fconfig=fconfig;
	ctx->notify=0;
	ctx->memwait=0;
	ctx->useshm=1;

	if((env=getenv("VGL_USEXSHM"))&&*env&&!strcmp(env, "0"))ctx->useshm=0;

	for(ctx->pool=ctx->head=ctx->tail=NULL,i=0;i<POOLSIZE;i++)
	{
		ctx->fbuffer[i].used=0;
		ctx->fbuffer[i].sync=0;
		ctx->fbuffer[i].next=ctx->pool;
		ctx->fbuffer[i].rf.opaque=&ctx->fbuffer[i];
		ctx->pool=&ctx->fbuffer[i];
		if(sem_init(&ctx->fbuffer[i].sem,0,0))
		{
			for(i--;i>=0;i--)sem_destroy(&ctx->fbuffer[i].sem);
			errmsg="RRTransInit: sem_init failure";
			goto fail2;
		}
	}

	if(sem_init(&ctx->sem,0,0))
	{
		errmsg="RRTransInit: sem_init failure";
		goto fail3;
	}

	if(sem_init(&ctx->msg,0,0))
	{
		errmsg="RRTransInit: sem_init failure";
		goto fail4;
	}

	if(sem_init(&ctx->mem,0,0))
	{
		errmsg="RRTransInit: sem_init failure";
		goto fail5;
	}

	if(pthread_mutex_init(&ctx->mtx,NULL))
	{
		errmsg="RRTransInit: pthread_mutex_init failure";
		goto fail6;
	}

#ifdef _GNU_SOURCE
	if(!(ctx->xlib=dlmopen(LM_ID_BASE,"libX11.so",RTLD_NOW)))
#else
#warning dlopen will not work with VirtualGL 2.3.1
	if(!(ctx->xlib=dlopen("libX11.so",RTLD_NOW)))
#endif
	{
		errmsg="RRTransInit: dlopen of libX11.so failed";
		goto fail7;
	}

	ctx->YCloseDisplay=dlsym(ctx->xlib,"XCloseDisplay");
	ctx->YCopyArea=dlsym(ctx->xlib,"XCopyArea");
	ctx->YCreateGC=dlsym(ctx->xlib,"XCreateGC");
	ctx->YCreateImage=dlsym(ctx->xlib,"XCreateImage");
	ctx->YCreatePixmap=dlsym(ctx->xlib,"XCreatePixmap");
	ctx->YDestroyImage=dlsym(ctx->xlib,"XDestroyImage");
	ctx->YFlush=dlsym(ctx->xlib,"XFlush");
	ctx->YFreeGC=dlsym(ctx->xlib,"XFreeGC");
	ctx->YFreePixmap=dlsym(ctx->xlib,"XFreePixmap");
	ctx->YGetWindowAttributes=dlsym(ctx->xlib,"XGetWindowAttributes");
	ctx->YLockDisplay=dlsym(ctx->xlib,"XLockDisplay");
	ctx->YOpenDisplay=dlsym(ctx->xlib,"XOpenDisplay");
	ctx->YPutImage=dlsym(ctx->xlib,"XPutImage");
	ctx->YSetErrorHandler=dlsym(ctx->xlib,"XSetErrorHandler");
	ctx->YSync=dlsym(ctx->xlib,"XSync");
	ctx->YUnlockDisplay=dlsym(ctx->xlib,"XUnlockDisplay");

	if(!ctx->YCloseDisplay||!ctx->YCopyArea||!ctx->YCreateGC||
		!ctx->YCreateImage||!ctx->YCreatePixmap||
		!ctx->YDestroyImage||!ctx->YFlush||!ctx->YFreeGC||
		!ctx->YFreePixmap||!ctx->YGetWindowAttributes||
		!ctx->YLockDisplay||!ctx->YOpenDisplay||
		!ctx->YPutImage||!ctx->YSetErrorHandler||
		!ctx->YUnlockDisplay||!ctx->YSync)
	{
		errmsg="RRTransInit: libX11.so is missing symbols";
		goto fail8;
	}

	if(!(ctx->xext=dlopen("libXext.so",RTLD_NOW)))
	{
		errmsg="RRTransInit: dlopen of libXext.so failed";
		goto fail8;
	}

	ctx->YShmAttach=dlsym(ctx->xext,"XShmAttach");
	ctx->YShmCreateImage=dlsym(ctx->xext,"XShmCreateImage");
	ctx->YShmDetach=dlsym(ctx->xext,"XShmDetach");
	ctx->YShmPutImage=dlsym(ctx->xext,"XShmPutImage");
	ctx->YShmQueryExtension=dlsym(ctx->xext,"XShmQueryExtension");

	if(!ctx->YShmAttach||!ctx->YShmCreateImage||!ctx->YShmDetach||
		!ctx->YShmPutImage||!ctx->YShmQueryExtension)
	{
		errmsg="RRTransInit: libXext.so is missing symbols";
		goto fail9;
	}

	if((ctx->format=getmode(ctx))==-1)
	{
		errmsg="RRTransInit: no usable graphics mode";
		goto fail9;
	}

	if(pthread_create(&ctx->h,NULL,worker,ctx))
	{
		errmsg="RRTransInit: pthread_create failure";
		goto fail9;
	}

	fconfig->compress=RRCOMP_PROXY;
	fconfig->gui=0;

	return ctx;

fail9:	dlclose(ctx->xext);
fail8:	dlclose(ctx->xlib);
fail7:	pthread_mutex_destroy(&ctx->mtx);
fail6:	sem_destroy(&ctx->mem);
fail5:	sem_destroy(&ctx->msg);
fail4:	sem_destroy(&ctx->sem);
fail3:	for(i=0;i<POOLSIZE;i++)sem_destroy(&ctx->fbuffer[i].sem);
fail2:	free(ctx);
fail1:	return NULL;
}

int RRTransDestroy(void *handle)
{
	int i;
	CTX *ctx=(CTX *)handle;

	pthread_cancel(ctx->h);
	pthread_join(ctx->h,NULL);

	dlclose(ctx->xext);
	dlclose(ctx->xlib);

	pthread_mutex_destroy(&ctx->mtx);

	sem_destroy(&ctx->mem);
	sem_destroy(&ctx->msg);
	sem_destroy(&ctx->sem);

	for(i=0;i<POOLSIZE;i++)
	{
		sem_destroy(&ctx->fbuffer[i].sem);

		if(ctx->fbuffer[i].used)
		{
			free(ctx->fbuffer[i].rf.bits);
			ctx->fbuffer[i].used=0;
		}
	}

	free(ctx);
	return 0;
}

int RRTransConnect(void *handle,char *receiver_name,int port)
{
	return 0;
}

int RRTransReady(void *handle)
{
	int r;
	CTX *ctx=(CTX *)handle;

	pthread_mutex_lock(&ctx->mtx);
	r=(ctx->head?0:1);
	pthread_mutex_unlock(&ctx->mtx);
	return r;
}

int RRTransSynchronize(void *handle)
{
	CTX *ctx=(CTX *)handle;

	while(1)
	{
		pthread_mutex_lock(&ctx->mtx);
		if(!ctx->head)
		{
			if(ctx->notify)
			{
				ctx->notify=0;
				pthread_mutex_unlock(&ctx->mtx);
				while(!sem_trywait(&ctx->msg));
			}
			else pthread_mutex_unlock(&ctx->mtx);
			return 0;
		}
		ctx->notify=1;
		pthread_mutex_unlock(&ctx->mtx);
		sem_wait(&ctx->msg);
	}
}

RRFrame *RRTransGetFrame(void *handle,int width,int height,int format,
	int stereo)
{
	FRAME *f;
	int psw;
	CTX *ctx=(CTX *)handle;

	if(format==RRTRANS_INDEX)
	{
		errmsg="RRTransGetFrame: RRTRANS_INDEX not supported";
		return NULL;
	}

	switch(rrtrans_ps[ctx->format])
	{
	case 1:	psw=width;
		break;
	case 2:	psw=width<<1;
		break;
	case 3:	psw=(width<<1)+width;
		break;
	case 4:	psw=width<<2;
		break;
	default:psw=width*rrtrans_ps[format];
		break;
	}

	psw=(psw+7)&0xfffffff8;

	while(1)
	{
		pthread_mutex_lock(&ctx->mtx);
		if((f=ctx->pool))
		{
			if(ctx->memwait)
			{
				ctx->memwait=0;
				ctx->pool=f->next;
				pthread_mutex_unlock(&ctx->mtx);
				while(!sem_trywait(&ctx->mem));
			}
			else
			{
				ctx->pool=f->next;
				pthread_mutex_unlock(&ctx->mtx);
			}
			break;
		}
		ctx->memwait=1;
		pthread_mutex_unlock(&ctx->mtx);
		sem_wait(&ctx->mem);
	}

	if(f->used)
	{
		if(f->rf.format==format&&f->rf.w==width&&f->rf.h==height)
			return &f->rf;

		free(f->rf.bits);
		f->used=0;
	}

	if(!(f->rf.bits=malloc(psw*height)))
	{
		pthread_mutex_lock(&ctx->mtx);
		f->next=ctx->pool;
		ctx->pool=f;
		pthread_mutex_unlock(&ctx->mtx);
		errmsg="RRTransGetFrame: malloc failure";
		return NULL;
	}

	f->rf.format=ctx->format;
	f->rf.w=width;
	f->rf.h=height;
	f->rf.pitch=psw;
	f->rf.rbits=NULL;
	f->used=1;

	return &f->rf;
}

int RRTransSendFrame(void *handle,RRFrame *frame,int sync)
{
	FRAME *f=(FRAME *)frame->opaque;
	CTX *ctx=(CTX *)handle;

	f->next=NULL;
	f->sync=sync;
	pthread_mutex_lock(&ctx->mtx);
	if(!ctx->head)ctx->head=ctx->tail=f;
	else
	{
		ctx->tail->next=f;
		ctx->tail=f;
	}
	sem_post(&ctx->sem);
	pthread_mutex_unlock(&ctx->mtx);
	if(sync)sem_wait(&f->sem);

	return 0;
}
